/******************************************************************************
 * Definisce le strutture dei registri ch servono (o dovrebbero servire) per
 * spegnere la CPU, riattivarla e monitorarla
 * ****************************************************************************/

#ifndef __CALZO_TEST_H__
#define __CALZO_TEST_H__

#define REG_BOOTROM_CODE_REG_GROUP	0xFFD080C0
#define REG_SYSTEM_MANAGER_MODULE	0xFFD08000
#define REG_RESET_MANAGER_MODULE	0xFFD05000

#define SOCFPGA_RSTMGR_CTRL	0x04
#define SOCFPGA_RSTMGR_MODMPURST	0x10
#define SOCFPGA_RSTMGR_MODPERRST	0x14
#define SOCFPGA_RSTMGR_BRGMODRST	0x1c

#define SOCFPGA_A10_RSTMGR_CTRL		0xC
#define SOCFPGA_A10_RSTMGR_MODMPURST	0x20
#define SOCFPGA_A10_RSTMGR_CTRL		0xC
#define SOCFPGA_A10_RSTMGR_PER0MODRST	0x24
#define SOCFPGA_A10_RSTMGR_PER1MODRST	0x28
#define SOCFPGA_A10_RSTMGR_BRGMODRST	0x2C

/* System Manager bits */
#define RSTMGR_CTRL_SWCOLDRSTREQ	0x1	/* Cold Reset */
#define RSTMGR_CTRL_SWWARMRSTREQ	0x2	/* Warm Reset */

#define RSTMGR_MPUMODRST_CPU1		0x2     /* CPU1 Reset */

#define SOCFPGA_REVISION_DEFAULT	0x1

/*MPU Module Reset Register */
#define RSTMGR_MPUMODRST_CPU0	0x1	/*CPU0 Reset*/
#define RSTMGR_MPUMODRST_CPU1	0x2	/*CPU1 Reset*/
#define RSTMGR_MPUMODRST_WDS		0x4	/*Watchdog Reset*/
#define RSTMGR_MPUMODRST_SCUPER	0x8	/*SCU and periphs reset*/
#define RSTMGR_MPUMODRST_L2		0x10	/*L2 Cache reset*/
#define SOCFPGA_ID_DEFAULT		0x1
#define SOCFPGA_REVISION_DEFAULT	0x1

#define SYSMGR_SILICON_ID1_OFFSET 0x0
#define SYSMGR_SILICON_ID1_REV_SHIFT 0
#define SYSMGR_SILICON_ID1_REV_MASK 0x0000FFFF
#define SYSMGR_SILICON_ID1_ID_SHIFT 16
#define SYSMGR_SILICON_ID1_ID_MASK 0xFFFF0000

#include <linux/types.h>

/**
 * System Manager Module Structure
 * */
typedef struct 
{
	__u32	stat;		// status register
	__u32	ctrl;
	__u32	counts;
	union {
		__u32	all;
		struct	{
			unsigned	cpu0	:1;	//00 credo che se scrivo 1 la CPU0 si resetta e resetta tutto il sistema (timer e periferiche comprese)
			unsigned	cpu1	:1;	//01 se settato a 0 la CPU1 parte. E' il contrario di CPU0
			unsigned	wds		:1;	//02 reset di entrambi gli status register WDog delle CPU
			unsigned	scuper	:1;	//03 reset di tutte le periferiche, compresi i timer di tutte le CPU
			unsigned	l2		:1;	//04 reset L2 controller
		} __attribute__((packed));
	} __attribute__((packed)) mpumodrst;
} __attribute__((packed)) reset_manager_module;

/**
 * Boot ROM Code Register Group
 * */
typedef struct 
{
	__u32	ctrl;		// Informazioni per controllare la BootRom
	__u32	cpu1startaddr;
} BootROM_code_register_group;

#endif	// __CALZO_TEST_H__