#include <linux/module.h>	/* Needed by all modules */
#include <linux/kernel.h>	/* Needed for KERN_INFO */
#include <linux/init.h>		/* Needed for the macros */

#include <linux/platform_device.h>
#include <linux/cpu.h>
#include <linux/of_irq.h>
#include <linux/of_address.h>	// per of_iomap()

#include <asm/io.h>	// per phys_to_virt
#include <asm/cacheflush.h>	// per flush_cache_all()

#include <asm/uaccess.h>
#include <linux/slab.h>	// per kmalloc & co
#include <linux/firmware.h>	// per request_firmware

//#include <../arch/arm/mach-socfpga/core.h>			// per l'accesso ai registri della CPU
#include "calzo_test01.h"

#define FIRMWARE_ADDR	0x20000000
#define FIRMWARE_MAX_SIZE	0x300000

// Da Core.h e socfpga.c
static void __iomem *sys_manager_base_addr;
static void __iomem *rst_manager_base_addr;
static void __iomem *sdr_ctl_base_addr;
static void __iomem *clkmgr_base_addr;
static unsigned int __iomem *cpu1start = NULL;

//unsigned long socfpga_cpu1start_addr = 0;
unsigned int cpu1start_addr;
extern unsigned int *__cpu1start_addr;	// punta alla fine del codice calzo_headsmp.S
extern unsigned int __start_bytecode, __end_bytecode;

/**
 * Configurazione dei puntatori ai registri si sistema letti dal device-tree.
 * Gli indirizzi fisici sono:
 * 	socfpga_cpu1start_addr =	0xFFD080C4
 * 	clkmgr_base_addr =			0xFFD04000
 * 	rst_manager_base_addr =		0xFFD05000
 * 	sdr_ctl_base_addr =			0xFFC25000
 * 	sys_manager_base_addr =		0xFFD08000
 * **/
void ct_sameas_socfpga_sysmgr_init(void)
{
	struct device_node *np;
	
	printk(KERN_INFO "Initialize CyconeV registers...\n");
	
	np = of_find_compatible_node(NULL, NULL, "altr,sys-mgr");
	
	if (of_property_read_u32(np, "cpu1-start-addr",
		(u32 *) &cpu1start_addr))
		pr_err("SMP: Need cpu1-start-addr in device tree.\n");
	sys_manager_base_addr = of_iomap(np, 0);
	
	np = of_find_compatible_node(NULL, NULL, "altr,rst-mgr");
	rst_manager_base_addr = of_iomap(np, 0);
	
	np = of_find_compatible_node(NULL, NULL, "altr,clk-mgr");
	clkmgr_base_addr = of_iomap(np, 0);
	
	np = of_find_compatible_node(NULL, NULL, "altr,sdr-ctl");
	sdr_ctl_base_addr = of_iomap(np, 0);
}

/**
 * Tenta di avviare la CPU1 con il baremetal!
 * See: arm/mach-socfpga/platsmp.c
 * **/
static int ct_sameas_socfpga_boot_secondary(__u32 jump_addr)
{
//	potrei linkare headsmp.o per avere i valori trampoline
#if 0
	int trampoline_size = &__start_bytecode - &__end_bytecode;
	
	if (cpu1start) {
		printk(KERN_INFO "Calzo: Prepare to JUMP\n");
		/// This will put CPU #1 into reset.
		writel(RSTMGR_MPUMODRST_CPU1,
			   rst_manager_base_addr + SOCFPGA_RSTMGR_MODMPURST);
		
		*__cpu1start_addr = cpu1start_addr;
		memcpy(phys_to_virt(0), &__start_bytecode, trampoline_size);
		*cpu1start = jump_addr;

		flush_cache_all();
		smp_wmb();
		outer_clean_range(0, trampoline_size);
		
		/// This will release CPU #1 out of reset.
		printk(KERN_INFO "Calzo: jump to %08X\n", *cpu1start);
		writel(0, rst_manager_base_addr + SOCFPGA_RSTMGR_MODMPURST);
	}
#else
	if (cpu1start) {
		printk(KERN_INFO "Calzo: Prepare to JUMP\n");
		/// This will put CPU #1 into reset.
		writel(RSTMGR_MPUMODRST_CPU1,
			rst_manager_base_addr + SOCFPGA_RSTMGR_MODMPURST);
		
		*cpu1start = jump_addr;
		
		flush_cache_all();
		smp_wmb();
		outer_clean_range(FIRMWARE_ADDR, FIRMWARE_MAX_SIZE);
		
		/// This will release CPU #1 out of reset.
		printk(KERN_INFO "Calzo: jump to %08X\n", *cpu1start);
		writel(0, rst_manager_base_addr + SOCFPGA_RSTMGR_MODMPURST);
	}
#endif
	
	return 0;
}

/**
 * ELF loader
 * Derivato da remoteproc_elf_loader.c, si vuole solo caricare la parte binaria del firmware
 * */
static int ct_elf_load_segment(const struct firmware *fw, void* virtual_fw_addr)
{
	struct elf32_hdr	*elf_header;
	struct elf32_phdr	*p_header;
	int i, ret = 0;
	const u8 *elf_data = fw->data;

	elf_header = (struct elf32_hdr*) elf_data;
	p_header = (struct elf32_phdr*) (elf_data + elf_header->e_phoff);
	
	/* Processa i vari segmenti dell'ELF */
	for(i=0; i<elf_header->e_phnum; i++, p_header++)
	{
		u32 da = p_header->p_paddr;
		u32 memsz = p_header->p_memsz;
		u32 filesz = p_header->p_filesz;
		u32 offset = p_header->p_offset;
		void *ptr;

		if (p_header->p_type != PT_LOAD)
			continue;
		
		printk(KERN_INFO "phdr: type %d da 0x%x memsz 0x%x filesz 0x%x\n", p_header->p_type, da, memsz, filesz);
		
		if (filesz > memsz) {
			printk(KERN_INFO "bad phdr filesz 0x%x memsz 0x%x\n", filesz, memsz);
			ret = -EINVAL;
			break;
		}
		
		if (offset + filesz > fw->size) {
			printk(KERN_INFO "truncated fw: need 0x%x avail 0x%zx\n", offset + filesz, fw->size);
			ret = -EINVAL;
			break;
		}
		
		/* grab the kernel address for this device address */
		ptr = (void*)virtual_fw_addr;
		
		/* put the segment where the remote processor expects it */
		if (p_header->p_filesz)
			memcpy(ptr, elf_data + p_header->p_offset, filesz);
		
		/*
		 * Zero out remaining memory for this segment.
		 *
		 * This isn't strictly required since dma_alloc_coherent already
		 * did this for us. albeit harmless, we may consider removing
		 * this.
		 */
		if (memsz > filesz)
			memset(ptr + filesz, 0, memsz - filesz);
	}
	
	return ret;
}


static unsigned char *MyFirmwareAddr;

const struct firmware *fw;

static int ct_probe(struct platform_device *pdev)
{
	const unsigned char *prop;	// property
	int i;

	printk(KERN_INFO "Calzo test: PROBE\n");
	ct_sameas_socfpga_sysmgr_init();
	
	// accesso ai registri di configurazione
	cpu1start = ioremap(cpu1start_addr, 0x4);

	// Print dei registri
	printk(KERN_INFO "socfpga_cpu1start_addr = 0x%08X\n", (unsigned)cpu1start_addr);
	printk(KERN_INFO "cpu1start = 0x%08X\n", (unsigned)cpu1start);
	printk(KERN_INFO "sys_manager_base_addr = 0x%08X\n", (unsigned)sys_manager_base_addr);
	printk(KERN_INFO "rst_manager_base_addr = 0x%08X\n", (unsigned)rst_manager_base_addr);
	printk(KERN_INFO "sdr_ctl_base_addr = 0x%08X\n", (unsigned)sdr_ctl_base_addr);
	
	printk(KERN_INFO "cpu1start value = 0x%08X\n", *(unsigned int*)cpu1start);
	printk(KERN_INFO "cpu1start value = 0x%08X\n", readl(cpu1start));
	
	printk(KERN_INFO "cpu_down(1) = %i\n", cpu_down(1));

	///	http://www.makelinux.net/ldd3/chp-9-sect-4	
	MyFirmwareAddr = (unsigned char *)phys_to_virt(FIRMWARE_ADDR);
	/// https://events.linuxfoundation.org/images/stories/pdf/lceu11_marinas.pdf
	/// http://lxr.free-electrons.com/source/mm/kmemleak.c#L909
	// dico al kernel che ho allocato un po' di ram
	kmemleak_alloc(MyFirmwareAddr, FIRMWARE_MAX_SIZE, 1, GFP_ATOMIC);
#if 0	
	printk(KERN_INFO "Physical Addr 0x%08X = Virtual 0x%08X\n", FIRMWARE_ADDR, (unsigned)MyFirmwareAddr);

	prop = of_get_property(pdev->dev.of_node, "firmware", NULL);
	if(prop)
	{
		printk(KERN_INFO "Firmware <%s> will be loaded.\n", prop);
		//https://www.kernel.org/doc/Documentation/firmware_class/README
		if(request_firmware(&fw, prop, NULL) < 0)
			printk(KERN_INFO "Error while loading firmware.\n");

		printk(KERN_INFO "Copy firmware into RAM\n");
		for(i = 0; i < fw->size; i++)
			MyFirmwareAddr[i] = fw->data[i];
		
		release_firmware(fw);
		fw = NULL;
	}
	else
		printk(KERN_INFO "Firmware property not found\n");
#else
	printk(KERN_INFO "Physical Addr 0x%08X = Virtual 0x%08X\n", FIRMWARE_ADDR, (unsigned)MyFirmwareAddr);
	
	prop = "bm.elf";	// baremetal in ELF format
	if(prop)
	{
		printk(KERN_INFO "Firmware <%s> will be loaded.\n", prop);
		//https://www.kernel.org/doc/Documentation/firmware_class/README
		if(request_firmware(&fw, prop, NULL) < 0)
		{
			printk(KERN_INFO "Error while loading firmware.\n");
			return -1;
		}
		
		printk(KERN_INFO "Copy firmware into RAM\n");
		ct_elf_load_segment(fw, (void*)MyFirmwareAddr);
		release_firmware(fw);
		fw = NULL;
	}
	else
		printk(KERN_INFO "Firmware property not found\n");
#endif
	
	//ct_sameas_socfpga_boot_secondary((__u32)MyFirmwareAddr);
	ct_sameas_socfpga_boot_secondary(FIRMWARE_ADDR);
	return 0;
}


static int ct_remove(struct platform_device *pdev)
{
	printk(KERN_INFO "Calzo test: REMOVE\n");
	cpu_up(1);
// Non so perchè, ma non devo liberarlo
//	kfree (MyFirmwareAddr);
	// tutto ciò che si alloca con kmemleak_ va liberato con kmemleak_ e non con kfree()
	kmemleak_free(MyFirmwareAddr);
	
	return 0;
}


static struct of_device_id calzo_test_match[] =
{
	{ .compatible = "altr,calzotest", },
	{ },
};

MODULE_DEVICE_TABLE(of, calzo_test_match);

static struct platform_driver calzo_test_diver =
{
	.probe = ct_probe,
	.remove = ct_remove,
	.driver = {
		.name = "calzo_test",
		.owner = THIS_MODULE,
		.of_match_table = calzo_test_match,
	},
};

module_platform_driver(calzo_test_diver);


MODULE_LICENSE("GPL v2");
MODULE_VERSION("0");
MODULE_AUTHOR("Calzo");
MODULE_DESCRIPTION("Descrizione del modulo");